#pragma once
#include "../GPU/gpu_types.hpp"
#include "../../math.h"

namespace JabyEngine {
    namespace GTE {
        namespace internal {
            template<typename T>
            struct VECTOR {
                T x;
                T y;
                T z;

                static constexpr VECTOR create() {
                    return VECTOR::create(0, 0, 0);
                }

                static constexpr VECTOR create(T x, T y, T z) {
                    return VECTOR{.x = x, .y = y, .z = z};
                }

                template<typename S>
                static constexpr VECTOR from(const GPU::Position<S>& pos) {
                    return VECTOR::create(static_cast<T>(pos.x), static_cast<T>(pos.y), 0);
                }

                template<typename S>
                constexpr S to() const {
                    return S::create(static_cast<S::PrimitiveType>(this->x), static_cast<S::PrimitiveType>(this->y));
                }
            };
        }
        using VECTOR  = internal::VECTOR<int32_t>;
        using SVECTOR = internal::VECTOR<int16_t>;

        struct ROTMATRIX {
            int16_t matrix[3][3];

            static constexpr ROTMATRIX identity() {
                return ROTMATRIX{.matrix = {
                        {4096, 0, 0},
                        {0, 4096, 0},
                        {0, 0, 4096}
                    }
                };
            } // TODO: replace int32_t with something weird for the DEG stuff??
            static ROTMATRIX rotated(int32_t x, int32_t y, int32_t z);
        };

        struct TRANSFERVECTOR : public VECTOR {
            static constexpr TRANSFERVECTOR identity() {
                return TRANSFERVECTOR::translated();
            }

            static constexpr TRANSFERVECTOR translated(int32_t x = 0, int32_t y = 0, int32_t z = 0) {
                return TRANSFERVECTOR{{.x = x, .y = y, .z = z}};
            }
        };
        
        struct MATRIX {
            ROTMATRIX      rotation;
            TRANSFERVECTOR transfer;            

            static constexpr MATRIX identity() {
                return MATRIX{.rotation = ROTMATRIX::identity(), .transfer = TRANSFERVECTOR::identity()};
            }
        };

        static constexpr auto one_degree       = FULL_CIRCLE/360;
        static constexpr auto one_tenth_degree = FULL_CIRCLE/3600;

        static constexpr int16_t in_degree(int32_t degree) {
            return degree*one_degree;
        }

        static constexpr int16_t in_degree10(int32_t degree) {
            return degree*one_tenth_degree;
        }
    }

    static constexpr unsigned long long operator""_DEG(unsigned long long degree) {
        return GTE::in_degree(degree);
    }

    static constexpr unsigned long long operator""_DEG10(unsigned long long degree) {
        return GTE::in_degree10(degree);
    }
}